from dataclasses import dataclass


@dataclass
class Conductor:
    resistivity: float
    temperature_coeff: float

    def get_resistivity(self, temperature: float):

        return self.resistivity * (1 + self.temperature_coeff * (temperature - 25))


@dataclass
class Ferrite:
    permeability: float
    max_flux_density: float
    losses: dict

    def get_loss_density(f: float, B: float, temperature: float = 25):
        raise NotImplementedError


# define some common materials as constants
COPPER = Conductor(1.68e-8, 0.0038)

N96 = Ferrite(
    permeability=2900,
    max_flux_density=500,
    losses={
        25: {
            13: {1: 2, 3: 4, 5: 6},
            25: {1: 2, 3: 4, 5: 6},
            50: {1: 2, 3: 4, 5: 6},
            100: {1: 2, 3: 4, 5: 6},
            200: {1: 2, 3: 4, 5: 6},
            300: {1: 2, 3: 4, 5: 6},
        },
        100: {
            13: {1: 2, 3: 4, 5: 6},
            25: {1: 2, 3: 4, 5: 6},
            50: {1: 2, 3: 4, 5: 6},
            100: {1: 2, 3: 4, 5: 6},
            200: {1: 2, 3: 4, 5: 6},
            300: {1: 2, 3: 4, 5: 6},
        },
    },
)



import math
from planar_magnetics.geometry import Arc, Point, Polygon
from planar_magnetics.kicad import Via
from planar_magnetics.utils import dcr_of_annulus
from planar_magnetics.windings.windings import Winding


class TopTurn(Winding):
    """Defines a top layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        termination_width: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # angle from "at" to the corner of the termination
        term_angle = math.asin(termination_width / outer_radius / 2)

        termination_arc = Arc(at, inner_radius, -term_angle, term_angle)

        # create the inner arc
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            term_angle,
            2 * math.pi - term_angle - inner_gap_angle - viastrip_angle,
        )

        via_arc = Arc(
            at,
            inner_radius,
            2 * math.pi - term_angle - inner_gap_angle - viastrip_angle,
            2 * math.pi - term_angle - inner_gap_angle,
        )

        # create outer arc
        outer_arc = Arc(
            at, outer_radius, 2 * math.pi - term_angle - outer_gap_angle, term_angle,
        )

        # create termination
        termination = [
            at + Point(outer_radius + termination_width, termination_width / 2),
            at + Point(outer_radius + termination_width, -termination_width / 2),
            at + Point(outer_radius * math.cos(term_angle), -termination_width / 2),
        ]

        # create the polygon
        points = [termination_arc, inner_arc, via_arc, outer_arc] + termination
        self.polygon = Polygon(points, layer)

    def __str__(self):

        return self.polygon.__str__()


class BottomTurn:
    """Defines a bottom layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        termination_width: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # angle from "at" to the corner of the termination
        term_angle = math.asin(termination_width / outer_radius / 2)

        termination_arc = Arc(at, inner_radius, term_angle, -term_angle)

        # create the inner arc
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            2 * math.pi - term_angle,
            term_angle + inner_gap_angle + viastrip_angle,
        )

        via_arc = Arc(
            at,
            inner_radius,
            term_angle + inner_gap_angle + viastrip_angle,
            term_angle + inner_gap_angle,
        )

        # create outer arc
        outer_arc = Arc(
            at, outer_radius, term_angle + outer_gap_angle, 2 * math.pi - term_angle,
        )

        # create termination
        termination = [
            at + Point(outer_radius + termination_width, -termination_width / 2),
            at + Point(outer_radius + termination_width, termination_width / 2),
            at + Point(outer_radius * math.cos(term_angle), termination_width / 2),
        ]

        # create the polygon
        points = [termination_arc, inner_arc, via_arc, outer_arc] + termination
        self.polygon = Polygon(points, layer)


class InnerTurn:
    """Defines a middle layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        rotation: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):
        assert (
            outer_radius > inner_radius
        ), "outer radius must be greater than inner radius"

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # create the arcs
        start_via_arc = Arc(
            at,
            inner_radius,
            inner_gap_angle / 2 + rotation,
            inner_gap_angle / 2 + viastrip_angle + rotation,
        )
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            inner_gap_angle / 2 + viastrip_angle + rotation,
            2 * math.pi - inner_gap_angle / 2 - viastrip_angle + rotation,
        )
        end_via_arc = Arc(
            at,
            inner_radius,
            2 * math.pi - inner_gap_angle / 2 - viastrip_angle + rotation,
            2 * math.pi - inner_gap_angle / 2 + rotation,
        )
        outer_arc = Arc(
            at,
            outer_radius,
            2 * math.pi - outer_gap_angle / 2 + rotation,
            outer_gap_angle / 2 + rotation,
        )

        # create the polygon
        points = [start_via_arc, inner_arc, end_via_arc, outer_arc]
        self.polygon = Polygon(points, layer)


class ViaStrip:
    def __init__(
        self,
        at: Point,
        layers: (str),
        inner_radius: float,
        start_angle: float,
        end_angle: float,
        size: float = 0.8,
        drill: float = 0.4,
    ):

        min_spacing = 0.5

        # calculate how may vias we can fit in the strip
        angle = end_angle - start_angle
        width = inner_radius * abs(angle)
        number_vias = int(width / (drill + min_spacing))

        # calculate via locations
        via_radius = inner_radius + size / 2
        delta_angle = angle / number_vias
        initial_angle = start_angle + delta_angle / 2
        angles = [initial_angle + n * delta_angle for n in range(number_vias)]
        locations = [
            Point(via_radius * math.cos(angle), via_radius * math.sin(angle))
            for angle in angles
        ]

        # create via strip
        self.vias = [Via(location + at, size, drill, layers) for location in locations]

    def __str__(self):
        expression = "\n".join([via.__str__() for via in self.vias])
        return expression


if __name__ == "__main__":
    turn = InnerTurn(Point(0, 0), 6, 12, 0.5, 0, math.pi / 8, 0, "F.Cu")
    print(turn)



import math
from planar_magnetics.geometry import Point
from planar_magnetics.windings import Spiral
from planar_magnetics.cores import Core
from planar_magnetics.kicad import Footprint, Reference, Value


class Transformer:
    """A multi-winding planar transformer

    Args:
        inner_radius: The inner radius of the windings in mm
        outer_radius: The outer radius of the windings in mm
        stackup: A dictionary of tuples specifying the layer and number of turns for each winding
        core_to_edge: The clearance required between the core surface and the pcb cutouts in 
        core_to_pcb: The distance between the core and the pcb surface in the z direction in mm
        board_thickness: The thickness of the PCB in mm
        trace_to_edge: The distance between winding edges and the edge cuts in mm
    """

    def __init__(
        self,
        inner_radius: float,
        outer_radius: float,
        stackup: dict,
        core_to_edge: float = 0.5,
        trace_to_edge: float = 1,
        core_to_pcb: float = 1,
        board_thickness: float = 1.6,
        opening_width: float = None,
        gap: float = 0,
    ):
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.stackup = stackup
        self.core_to_edge = core_to_edge
        self.trace_to_edge = trace_to_edge
        self.core_to_pcb = core_to_pcb
        self.board_thickness = board_thickness
        if opening_width is None:
            self.opening_width = outer_radius - inner_radius
        else:
            self.opening_width = opening_width

        # create the layers
        self.layers = [
            Spiral(inner_radius, outer_radius, n, s, l) for l, n, s in stackup
        ]

        # create the core
        core_to_trace = core_to_edge + trace_to_edge
        self.core = Core(
            centerpost_radius=inner_radius - core_to_trace,
            window_width=(outer_radius - inner_radius) + 2 * core_to_trace,
            window_height=board_thickness + 2 * core_to_pcb,
            opening_width=self.opening_width + 2 * core_to_trace,
            gap=gap,
        )

    def to_kicad_footprint(
        self,
        name: str,
        create_core_step: bool = False,
        freecad_path: str = "C:/Program Files/FreeCAD 0.19/bin",
    ):
        """Export the Cffc inductor design as a KiCAD footprint file (*.kicad_mods)
        """

        # add the reference and value silkscreens
        x_loc = self.core.width / 2 + 1
        height_avail = (self.core.width - self.opening_width) / 2
        font_size = min(2, height_avail / 4)
        val_loc = Point(x_loc, self.opening_width / 2 + height_avail / 3)
        ref_loc = Point(x_loc, self.opening_width / 2 + 2 * height_avail / 3)
        reference = Reference(ref_loc, font_size)
        value = Value(val_loc, font_size)

        cutouts = self.core.create_pcb_cutouts(Point(0, 0), self.core_to_edge)

        # create a footprint from the various elements
        contents = cutouts + self.layers + [reference, value]
        footprint = Footprint(name, contents=contents)

        # write the footprint to a file
        fh = open(f"{name}.kicad_mod", "w")
        fh.write(footprint.__str__())
        fh.close()

        if create_core_step:
            self.core.to_step(f"{name}.step", self.core_to_pcb - 0.1, 0.1, freecad_path)

    def plot(self):
        import matplotlib.pyplot as mp

        n_rows = min(2, len(self.layers))
        n_columns = math.ceil(len(self.layers) / 2)
        fig, axes = mp.subplots(n_rows, n_columns)

        for layer, ax in zip(self.layers, axes.ravel()):
            layer.plot(ax=ax)


if __name__ == "__main__":

    transformer = Transformer(
        inner_radius=3,
        outer_radius=6,
        stackup=[
            ("In1.Cu", 1, 0.1),
            ("In1.Cu", 5, 0.1),
            ("In2.Cu", 5, 0.1),
            ("In1.Cu", 1, 0.1),
        ],
    )

    import matplotlib.pyplot as mp

    transformer.plot()
    mp.show()



import math
from typing import Union
from pathlib import Path


class Winding:
    def rotate(self, angle: float):
        """Rotate a winding about its center"""
        rotated = self
        rotated.polygon = self.polygon.rotate_about(self.at, angle)
        return rotated

    def plot(self, ax=None, max_angle: float = math.pi / 36):
        self.polygon.plot(ax, max_angle)

    def to_dxf(
        self,
        filename: Union[str, Path],
        version: str = "R2000",
        encoding: str = None,
        fmt: str = "asc",
    ) -> None:
        """Export the polygon to a dxf file

        Args:
            filename: file name as string
            version: DXF version
            encoding: override default encoding as Python encoding string like ``'utf-8'``
            fmt: ``'asc'`` for ASCII DXF (default) or ``'bin'`` for Binary DXF
        """

        self.polygon.to_dxf(filename, version, encoding, fmt)

    def __str__(self):
        """Print KiCAD S-Expression of the spiral.  Assumes units are mm."""
        return self.polygon.__str__()
